"""
Package: src
Filename: auth.py
Author(s): Grant W
"""
# Python Imports
import hashlib
import hmac
from datetime import datetime, timedelta
import json
from typing import Union
from urllib.parse import urlencode

# Third Party Imports

# Local Imports
from .logs import logger
from .helpers import make_session
from .exceptions import HttpTimeoutException, RexServiceException


class Signer:
    """
    Create signatures for requests to the Rex Registry API and Gateway
    """
    def __init__(self,
                 access_key: str,
                 secret_key: str,
                 signing_string: str = '',
                 default_ttl: int = 10):
        self.access_key = access_key
        self.secret_key = secret_key
        self.signing_string = signing_string
        self.default_ttl = default_ttl

    def is_expired(self, expiration_date_str: str) -> bool:
        """Check if the expiration date is in the past
        """
        expiration_date = datetime.strptime(expiration_date_str,
                                            '%Y-%m-%dT%H:%M:%S.%fZ')
        return expiration_date < datetime.utcnow()

    @staticmethod
    def digest(signing_key: str, message: str, encoding='utf-8') -> str:
        """
        Create a digest from signing_key & message
        """
        _hmac = hmac.new(
            key=bytes(signing_key, encoding=encoding),
            msg=bytes(message, encoding=encoding),
            digestmod=hashlib.sha256
        )
        return _hmac.hexdigest()

    @staticmethod
    def create_expiration_date(ttl: int) -> str:
        """
        Create a formatted date string from now + ttl in seconds
        - expected format:         "2021-03-30T14:17:29.208Z"
        - python isoformat outputs '2021-03-30T14:17:29.208292', so we have to
        strip the last 3, and append the Z.
        :ttl : int - seconds
        """
        return (datetime.utcnow() + timedelta(seconds=ttl)) \
            .isoformat(timespec="milliseconds") + "Z"

    @classmethod
    def sign_from_ttl(cls,
                      access_key,
                      secret_key,
                      ttl,
                      signing_string: str = '') -> (str, str):
        expiration_date_str = cls.create_expiration_date(ttl)
        first = cls.digest(expiration_date_str, signing_string)
        second = cls.digest(access_key, first)
        final = cls.digest(secret_key, second)
        return final, expiration_date_str

    @classmethod
    def sign_from_expiration_date(cls,
                                  access_key,
                                  secret_key,
                                  expiration_date_str,
                                  signing_string: str = ''
                                  ) -> (str, str):
        first = cls.digest(expiration_date_str, signing_string)
        second = cls.digest(access_key, first)
        final = cls.digest(secret_key, second)
        return final, expiration_date_str

    def _create_query_params_signing_string(self, parameters: dict) -> str:
        """Create a signing string for use in Gateway from query parameters
        """
        if 'signing_string' in parameters:
            del parameters['signing_string']
        if 'signature' in parameters:
            del parameters['signature']
        sorted_params = sorted(parameters.items())
        encoded_params = urlencode(sorted_params)
        return hashlib.sha256(encoded_params.encode('utf-8')).hexdigest()

    def _create_request_body_signing_string(self, request_body: str) -> str:
        """SHA256 digest of the request body as a hexidecimal string
        in lowercase
        """
        return hashlib.sha256(request_body.encode('utf-8')).hexdigest()

    def sign_query_params_from_expiration_date(self,
                                               parameters: dict,
                                               expiration_date_str: str,
                                               access_key: str = None,
                                               secret_key: str = None,
                                               as_dict=False) -> str:
        if access_key is None:
            access_key = self.access_key
        if secret_key is None:
            secret_key = self.secret_key
        parameters['access_key'] = access_key
        parameters['expiration'] = expiration_date_str
        signing_string = self._create_query_params_signing_string(parameters)
        parameters['signing_string'] = signing_string
        signature, _ = self.sign_from_expiration_date(access_key,
                                                      secret_key,
                                                      expiration_date_str,
                                                      signing_string)
        parameters['signature'] = signature

        # Remove signing string for smrg
        del parameters['signing_string']
        if as_dict:
            return parameters
        return urlencode(parameters, doseq=True)

    def sign_query_parameters_from_expiration_date(self, *args, **kwargs):
        return self.sign_query_params_from_expiration_date(*args, **kwargs)

    def sign_query_params_from_ttl(self, *args, **kwargs):
        return self.sign_query_parameters_from_ttl(*args, **kwargs)

    def sign_query_parameters_from_ttl(self,
                                       parameters: dict,
                                       ttl: Union[int, None] = None,
                                       access_key: str = None,
                                       secret_key: str = None,
                                       as_dict=False) -> str:
        if ttl is None:
            ttl = self.ttl
        expiration_date_str = self.create_expiration_date(ttl)
        return self.sign_query_params_from_expiration_date(parameters,
                                                           expiration_date_str,
                                                           access_key,
                                                           secret_key,
                                                           as_dict)


class RexRequest:
    """Wrapper for http calls to include our signature"""

    def __init__(self,
                 access_key,
                 secret_key,
                 default_ttl: int = 10):
        self.default_ttl = default_ttl
        self.access_key = access_key
        self.secret_key = secret_key
        self.signer = Signer(access_key, secret_key)
        self.session = make_session()

    def _signature(self, ttl: int = None, signing_string: str = None) -> str:
        if ttl is None:
            ttl = self.default_ttl
        return self.signer.sign_from_ttl(
            self.access_key,
            self.secret_key,
            ttl,
            signing_string=signing_string
        )

    def _create_auth_headers(self,
                             additional_headers={},
                             body=''):
        signing_string = self.signer._create_request_body_signing_string(body)

        signature, expiration = self._signature(signing_string=signing_string)
        base = {
            'dynata-expiration': expiration,
            'dynata-access-key': self.access_key,
            'dynata-signature': signature,
            'dynata-signing-string': signing_string
        }
        return dict(additional_headers, **base)

    def dispatch(self,
                 url,
                 data='',
                 method='GET') -> Union[dict, str]:

        additional_headers = {}
        if data:
            additional_headers = {'Content-type': 'application/json'}
            data = json.dumps(data)

        headers = self._create_auth_headers(additional_headers,
                                            body=data)

        if not hasattr(self.session, method.lower()):
            raise AttributeError('Invalid http method provided.')

        method = getattr(self.session, method.lower())

        res = method(url, data=data, headers=headers)
        if res.status_code > 299:
            if res.status_code == 504:
                raise HttpTimeoutException(res.content)
            if data:
                logger.warning(data)
            logger.warning(res.__dict__)
            raise RexServiceException(res.content.decode('utf-8'))
        try:
            return res.json()
        except json.decoder.JSONDecodeError as e:
            if res.status_code != 204:
                logger.debug(str(e), res.__dict__, exc_info=1)
        return res.content.decode('utf-8')

    def get(self, url: str):
        return self.dispatch(url)

    def post(self, url, data):
        return self.dispatch(url, data=data, method='POST')



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

context_id = 'super-unique-ctx-id'

context = gateway.get_context(context_id)

# {
#    'id': 'super-unique-ctx-id',
#    'items': {
#        'ctx': 'parent-context-id',
#        'gender': 'male',
#        'birth_data': '1999-09-09',
#        'postal_code': '90210'
#     }
# }



"""
Package: dynata-rex.models
Filename: respondent_gateway.py
Author(s): Grant W

Description: Implementation of dataclasses for Respondent Gateway
"""
# Python Imports
from enum import Enum


class GatewayGenderEnum(Enum):
    MALE = 1
    FEMALE = 2


class GatewayDispositionsEnum(Enum):
    UNKNOWN = 0
    COMPLETE = 1
    TERMINATION = 2
    OVERQUOTA = 3
    DUPLICATE = 4
    QUALITY = 5


class GatewayStatusEnum(Enum):
    """
    Composite status code from Disposition + non-unique status code
    Must call as tuple of Gateway Disposition + Status code
    ie
    >>> disp_code, status_code = 1, 0
    >>> disp = GatewayDispositionsEnum(disp_code)
    >>> status = RespondentGatewayStatusEnum((disp, status_code))
    >>> status.name
    'COMPLETE_DEFAULT`
    >>> status.value
    (<GatewayDispositionsEnum.COMPLETE: 1>, 0)
    """
    UNKNOWN_DEFAULT = GatewayDispositionsEnum.UNKNOWN, 0
    COMPLETE_DEFAULT = GatewayDispositionsEnum.COMPLETE, 0
    COMPLETE_PARTIAL = GatewayDispositionsEnum.COMPLETE, 1
    TERMINATION_DYNATA = GatewayDispositionsEnum.TERMINATION, 1
    TERMINATION_CLIENT = GatewayDispositionsEnum.TERMINATION, 2
    OVERQUOTA_DYNATA = GatewayDispositionsEnum.OVERQUOTA, 1
    OVERQUOTA_CLIENT = GatewayDispositionsEnum.OVERQUOTA, 2
    DUPLICATE_DEFAULT = GatewayDispositionsEnum.DUPLICATE, 0
    QUALITY_ANSWER = GatewayDispositionsEnum.QUALITY, 1
    QUALITY_SPEEDING = GatewayDispositionsEnum.QUALITY, 2
    QUALITY_SUSPENDED = GatewayDispositionsEnum.QUALITY, 3



"""
Package: src.src
Filename: logs
Author(s): Grant W, Nick P

Description: logging
"""
# Python Imports
import logging
import os

# Third Party Imports

# Local Imports

LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO').upper()


def create_logger(level):
    logger = logging.getLogger(__name__)
    logger.setLevel(level)
    return logger


logger = create_logger(LOG_LEVEL)



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

context_id = 'super-unique-ctx-id'

data = {
    'ctx': 'parent-context-id',           # From survey link 'ctx' parameter
    'gender': 'male',
    'birth_data': '1999-09-09',
    'postal_code': '90210'
}

gateway.create_context(context_id, data)



"""
Package: src.src
Filename: helpers.py
Author(s): Grant W

Description: General helpers
"""
# Python Imports
import os
import random
import time

# Third Party Imports
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ReadTimeout, ConnectionError

# Local Imports
from .exceptions import HttpTimeoutException, RexServiceException


DEFAULT_TIMEOUT = int(os.environ.get('DEFAULT_TIMEOUT', '60'))
DEFAULT_RETRIES = int(os.environ.get('DEFAULT_RETRIES', '3'))


class TimeoutHTTPAdapter(HTTPAdapter):

    def _wait(self, try_number):
        interval = (2 ** try_number) + (random.randint(0, 1000) / 1000)
        time.sleep(interval)
        return

    def __init__(self, *args, **kwargs):
        _kwargs = {k: v for k, v in kwargs.items() if k != 'request_timeout'}
        super().__init__(*args, **_kwargs)
        self.maximum_retries = DEFAULT_RETRIES

    def send(self, request, try_count=1, **kwargs):
        if "request_timeout" in kwargs:
            request_timeout = kwargs["request_timeout"]
            del kwargs["request_timeout"]
        else:
            request_timeout = DEFAULT_TIMEOUT
        kwargs["timeout"] = request_timeout
        try:
            return super().send(request, **kwargs)
        except ReadTimeout as e:
            raise HttpTimeoutException(e)
        except ConnectionError as e:
            try_count += 1
            if try_count > self.maximum_retries:
                raise RexServiceException from e
            self._wait(try_count)
            return self.send(request, try_count=try_count, **kwargs)


def make_session(request_timeout=DEFAULT_TIMEOUT):
    """Make a session and mount the TimeoutHTTPAdapter"""
    session = requests.Session()
    adapter = TimeoutHTTPAdapter(request_timeout=request_timeout)
    # TODO: Set dynata_rex.__version__ and use instead of 0.0.1
    session.headers["User-Agent"] = 'rex-sdk-python/0.0.1'
    session.mount('https://', adapter)
    session.mount('http://', adapter)
    return session



from dynata_rex import RespondentGateway

gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

url = 'https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en'

signed_url = gateway.sign_url(url)

# "https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en&access_key=rex_access_key&expiration=2021-11-24T16:12:06.070Z&signature=fa8b5cac82d34bcf8026904b  # noqa



"""
Package: src.models
Filename: base.py
Author(s): Grant W

Description: Base models for REX sdk
"""
# Python Imports
import hashlib
from enum import Enum

# Third Party Imports
from pydantic import BaseModel

# Local Imports


class BaseObject:
    def __str__(self):
        return f"<dynata_rex.{self.__class__.__name__}>"


class HashableModel(BaseObject, BaseModel):
    """
    Base class for all objects in the Opportunity Registry
    """
    def __hash__(self):
        try:
            return hash((type(self),) + tuple(self.__dict__.values()))
        except TypeError:
            _json = self.json(sort_keys=True)
            to_hash = str(type(self)) + _json
            digest = hashlib.sha256(to_hash.encode('utf-8')).hexdigest()
            return int(digest, 16) % 10**19


class FallbackEnum(Enum):
    """Override missing prop to return a default if defined"""

    @classmethod
    def _missing_(cls, value):
        if hasattr(cls, 'DEFAULT'):
            return getattr(cls, 'DEFAULT')
        raise ValueError(f"Enum {cls.__name__} does not have a defined "
                         "default value")



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

url = 'https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en'
date_of_birth = '1990-01-01'
gender = 'male'
postal_code = '90210'
respondent_id = 'very-unique-respondent-id'
ttl = 10  # in seconds

signed_link = gateway.create_respondent_url(url,
                                            date_of_birth,
                                            gender,
                                            postal_code,
                                            respondent_id,
                                            ttl=ttl)

# https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en&birth_date=1990-01-01&gender=male&postal_code=90210&respondent_id=very-unique-respondent-id&access_key=rex_access_key&expiration=2021-11-29T15:35:12.993Z&signature=4353e8c4ca8f8fb75530214ac78139b55ca3f090438c639476b8584afe1396e6 # noqa


from dynata_rex import OpportunityRegistry
registry = OpportunityRegistry('rex_access_key', 'rex_secret_key')


opportunities = registry.list_opportunities()

opportunity = opportunities[0]

# Get a list of corresponding opportunities from a project_id
corresponding = registry.list_project_opportunities(opportunity.project_id)
